import numpy as np
from copy import deepcopy

from base.base import BaseEstimator, ClassifierMixin, RegressorMixin


# TODO : revise the code
class VotingClassifier(ClassifierMixin, BaseEstimator):
    def __init__(self, estimators, voting='hard'):
        self.estimators = estimators
        self.voting = voting

    def fit(self, X, y):
        self.classes_, y_train = np.unique(y, return_inverse=True)
        self.estimators_ = [deepcopy(est).fit(X, y_train) for _, est in self.estimators]
        self._is_fitted = True
        return self

    def transform(self, X):
        if self.voting == 'hard':
            prob = np.array([est.predict(X) for est in self.estimators_]).T
        elif self.voting == 'soft':
            prob = np.array([est.predict_proba(X) for est in self.estimators_])
        return prob

    def predict(self, X):
        prob = self.transform(X)
        if self.voting == 'hard':
            pred = np.apply_along_axis(lambda x:np.argmax(np.bincount(x)), axis=1, arr=prob)
        elif self.voting == 'soft':
            pred = np.argmax(np.mean(prob, axis=0), axis=1)
        return self.classes_[pred]

    def predict_proba(self, X):
        if self.voting == 'hard':
            raise AttributeError
        return np.mean(self.transform(X), axis=0)
    

class VotingRegressor(RegressorMixin, BaseEstimator):
    
    def __init__(self, estimators):
        self.estimators = estimators

    def fit(self, X, y):
        self.estimators_ = [deepcopy(est).fit(X, y) for _, est in self.estimators]
        self._is_fitted = True
        return self

    def transform(self, X):
        prob = np.array([est.predict(X) for est in self.estimators_]).T
        return prob

    def predict(self, X):
        prob = self.transform(X)
        return np.mean(prob, axis=1)
    




from utils.__clone import clone
import numpy as np


from model_selection.__split import KFold, StratifiedKFold



def cross_val_predict(estimator, X, y, method="predict"):
    if estimator._estimator_type == "regressor":
        cv = KFold()
    else:  # estimator._estimator_type == "classifier"
        cv = StratifiedKFold()
    predictions = []
    indices = []
    for train, test in cv.split(X, y):
        est = clone(estimator)
        est.fit(X[train], y[train])
        predictions.extend(getattr(est, method)(X[test]))
        indices.extend(test)
    inv_indices = np.empty(len(indices), dtype=int)
    inv_indices[indices] = np.arange(len(indices))
    return np.array(predictions)[inv_indices]



#TODO : revise the code 

import numpy as np

from abc import ABC, abstractmethod


class BaseKFold(ABC):
    def __init__(self, n_splits=5, shuffle=False, random_state=0) -> None:
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_state = random_state

    def split(self, X, y):
        indices = np.arange(X.shape[0])
        for test_index in self._iter_test_masks(X, y):
            yield indices[~test_index], indices[test_index]

    @abstractmethod
    def _iter_test_masks(self, X, y):
        pass 

class KFold(BaseKFold):

    def __init__(self, n_splits=5, shuffle=False, random_state=0) -> None:
        super().__init__(n_splits, shuffle, random_state)

    def _iter_test_indices(self, X, y):
        indices = np.arange(X.shape[0])
        if self.shuffle:
            rng = np.random.RandomState(self.random_state)
            rng.shuffle(indices)
        fold_sizes = np.full(self.n_splits, X.shape[0] // self.n_splits)
        fold_sizes[:X.shape[0] % self.n_splits] += 1
        current = 0
        for fold_size in fold_sizes:
            yield indices[current:current + fold_size]
            current += fold_size

    def _iter_test_masks(self, X, y):
        for test_index in self._iter_test_indices(X, y):
            test_mask = np.zeros(X.shape[0], dtype=bool)
            test_mask[test_index] = True
            yield test_mask

    def split(self, X, y):
        return super().split(X, y)


class StratifiedKFold(BaseKFold):

    def __init__(self, n_splits=5, shuffle=False, random_state=0) -> None:
        super().__init__(n_splits, shuffle, random_state)

    def _kfold(self, count, rng):
        indices = np.arange(count)
        if self.shuffle:
            rng.shuffle(indices)
        fold_sizes = np.full(self.n_splits, count // self.n_splits)
        fold_sizes[:count % self.n_splits] += 1
        current = 0
        for fold_size in fold_sizes:
            test_mask = np.zeros(count, dtype=bool)
            test_mask[current:current + fold_size] = True
            yield indices[test_mask]
            current += fold_size

    def _make_test_folds(self, X, y):
        rng = np.random.RandomState(self.random_state)
        unique_y, y_inversed = np.unique(y, return_inverse=True)
        y_counts = np.bincount(y_inversed)
        test_folds = np.zeros(X.shape[0])
        per_cls_cvs = [self._kfold(count, rng) for count in y_counts]
        test_folds = np.zeros(X.shape[0])
        for test_fold_indices, per_cls_splits in enumerate(zip(*per_cls_cvs)):
            for cls, test_split in zip(unique_y, per_cls_splits):
                cls_test_folds = test_folds[y == cls]
                cls_test_folds[test_split] = test_fold_indices
                test_folds[y == cls] = cls_test_folds
        return test_folds

    def _iter_test_masks(self, X, y):
        test_folds = self._make_test_folds(X, y)
        for i in range(self.n_splits):
            yield test_folds == i

    def split(self, X, y):
        return super().split(X, y)
    

def train_test_split(*arrays, test_size=0.25, random_state=None, shuffle=True):
    """
    Split arrays or matrices into random train and test subsets.

    Parameters:
        *arrays : sequence of indexables with same length / shape[0]
            Allowed inputs are lists, numpy arrays, scipy-sparse
            matrices or pandas dataframes.
        test_size : float or int, default=0.25
            If float, should be between 0.0 and 1.0 and represent the
            proportion of the dataset to include in the test split. If
            int, represents the absolute number of test samples.
        random_state : int or None, default=None
            Controls the shuffling applied to the data before applying the split.
            Pass an int for reproducible output across multiple function calls.
        shuffle : bool, default=True
            Whether or not to shuffle the data before splitting. If shuffle=False
            then stratify must be None.

    Returns:
        splitting : list, length=2 * len(arrays)
            List containing train-test split of inputs.
    """
    n_samples = len(arrays[0])
    if random_state is not None:
        np.random.seed(random_state)

    indices = np.arange(n_samples)
    if shuffle:
        np.random.shuffle(indices)

    if isinstance(test_size, float):
        test_size = int(n_samples * test_size)

    test_indices = indices[:test_size]
    train_indices = indices[test_size:]

    train_arrays = [np.take(arr, train_indices, axis=0) for arr in arrays]
    test_arrays = [np.take(arr, test_indices, axis=0) for arr in arrays]

    return [*train_arrays, *test_arrays]


import numpy as np
import numbers
import math


def _unique(values, *, return_inverse=False, return_counts=False):
    """Helper function to find unique values with support for python objects."""
    if values.dtype == object:
        return _unique_python(values, return_inverse=return_inverse, return_counts=return_counts)
    # numerical
    return _unique_np(values, return_inverse=return_inverse, return_counts=return_counts)


def _unique_np(values, return_inverse=False, return_counts=False):
    """Helper function to find unique values for numpy arrays that correctly
    accounts for nans. See `_unique` documentation for details."""
    uniques = np.unique(
        values, return_inverse=return_inverse, return_counts=return_counts
    )

    inverse, counts = None, None

    if return_counts:
        *uniques, counts = uniques

    if return_inverse:
        *uniques, inverse = uniques

    if return_counts or return_inverse:
        uniques = uniques[0]

    # np.unique will have duplicate missing values at the end of `uniques`
    # here we clip the nans and remove it from uniques
    if uniques.size and is_scalar_nan(uniques[-1]):
        nan_idx = np.searchsorted(uniques, np.nan)
        uniques = uniques[: nan_idx + 1]
        if return_inverse:
            inverse[inverse > nan_idx] = nan_idx

        if return_counts:
            counts[nan_idx] = np.sum(counts[nan_idx:])
            counts = counts[: nan_idx + 1]

    ret = (uniques,)

    if return_inverse:
        ret += (inverse,)

    if return_counts:
        ret += (counts,)

    return ret[0] if len(ret) == 1 else ret


def _unique_python(values, *, return_inverse, return_counts):
    # Only used in `_uniques`, see docstring there for details
    try:
        uniques_set = set(values)
        uniques_set, missing_values = _extract_missing(uniques_set)

        uniques = sorted(uniques_set)
        uniques.extend(missing_values.to_list())
        uniques = np.array(uniques, dtype=values.dtype)
    except TypeError:
        types = sorted(t.__qualname__ for t in set(type(v) for v in values))
        raise TypeError(
            "Encoders require their input argument must be uniformly "
            f"strings or numbers. Got {types}"
        )
    ret = (uniques,)

    if return_inverse:
        ret += (_map_to_integer(values, uniques),)

    if return_counts:
        ret += (_get_counts(values, uniques),)

    return ret[0] if len(ret) == 1 else ret




def _encode(values, *, uniques, check_unknown=True):
    """Helper function to encode values into [0, n_uniques - 1].

    Uses pure python method for object dtype, and numpy method for
    all other dtypes.
    The numpy method has the limitation that the `uniques` need to
    be sorted. Importantly, this is not checked but assumed to already be
    the case. The calling method needs to ensure this for all non-object
    values.

    Parameters
    ----------
    values : ndarray
        Values to encode.
    uniques : ndarray
        The unique values in `values`. If the dtype is not object, then
        `uniques` needs to be sorted.
    check_unknown : bool, default=True
        If True, check for values in `values` that are not in `unique`
        and raise an error. This is ignored for object dtype, and treated as
        True in this case. This parameter is useful for
        _BaseEncoder._transform() to avoid calling _check_unknown()
        twice.

    Returns
    -------
    encoded : ndarray
        Encoded values
    """
    if values.dtype.kind in "OUS":
        try:
            return _map_to_integer(values, uniques)
        except KeyError as e:
            raise ValueError(f"y contains previously unseen labels: {str(e)}")
    else:
        if check_unknown:
            diff = _check_unknown(values, uniques)
            if diff:
                raise ValueError(f"y contains previously unseen labels: {str(diff)}")
        return np.searchsorted(uniques, values)


def _map_to_integer(values, uniques):
    """Map values to integers based on a set of unique values."""
    return {val: i for i, val in enumerate(uniques)}[values]


def _check_unknown(values, uniques):
    """Check for unknown values in `values`."""
    return set(values) - set(uniques)



def _get_counts(values, uniques):
    """Get the count of each unique value in values."""
    if values.dtype.kind in "OU":
        return _get_counts_object(values, uniques)
    else:
        return _get_counts_numeric(values, uniques)


def _get_counts_object(values, uniques):
    """Get the count of each unique object value in values."""
    output = np.zeros(len(uniques), dtype=np.int64)
    for i, item in enumerate(uniques):
        output[i] = np.count_nonzero(values == item)
    return output


def _get_counts_numeric(values, uniques):
    """Get the count of each unique numeric value in values."""
    unique_values, counts = np.unique(values, return_counts=True)
    index_mapping = {val: i for i, val in enumerate(unique_values)}
    output = np.zeros_like(uniques, dtype=np.int64)
    for i, item in enumerate(uniques):
        if item in index_mapping:
            output[i] = counts[index_mapping[item]]
    return output


class MissingValues:
    def __init__(self, nan=False, none=False):
        self.nan = nan
        self.none = none

def _extract_missing(values):
    """Extract missing values from a set."""
    missing_values_set = {value for value in values if value is None or is_scalar_nan(value)}

    nan_present = np.nan in missing_values_set
    none_present = None in missing_values_set

    if not missing_values_set:
        return values, MissingValues(nan=False, none=False)

    return values - missing_values_set, MissingValues(nan=nan_present, none=none_present)


def is_scalar_nan(x):
    """Test if x is NaN.

    Parameters
    ----------
    x : any type
        Any scalar value.

    Returns
    -------
    bool
        Returns true if x is NaN, and false otherwise.
    """
    return (
        not isinstance(x, numbers.Integral)
        and isinstance(x, numbers.Real)
        and math.isnan(x)
    )


import numpy as np

def roc_curve(y_true, y_score):
    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_true = y_true[desc_score_indices]
    y_score = y_score[desc_score_indices]
    distinct_value_indices = np.where(np.diff(y_score))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]
    tps = np.cumsum(y_true)[threshold_idxs]
    fps = 1 + threshold_idxs - tps
    thresholds = y_score[threshold_idxs]
    tps = np.r_[0, tps]
    fps = np.r_[0, fps]
    thresholds = np.r_[thresholds[0] + 1, thresholds]
    tpr = tps / tps[-1]
    fpr = fps / fps[-1]
    return fpr, tpr, thresholds





import numpy as np
from base.base import BaseEstimator, ClassifierMixin
from tree.decision_tree import DecisionTreeClassifier, DecisionTreeRegressor
from scipy.special import expit, logsumexp


class AdaBoostClassifier(ClassifierMixin, BaseEstimator):
    def __init__(self, n_estimators=50, random_state=0):
        self.n_estimators = n_estimators
        self.random_state = random_state

    def fit(self, X, y):
        self.classes_, y_train = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)
        sample_weight = np.full(X.shape[0], 1 / X.shape[0])
        self.estimators_ = []
        self.estimator_weights_ = np.zeros(self.n_estimators)
        self.estimator_errors_ = np.ones(self.n_estimators)
        MAX_INT = np.iinfo(np.int32).max
        rng = np.random.RandomState(self.random_state)
        for i in range(self.n_estimators):
            est = DecisionTreeClassifier(max_depth=1)
            est.fit(X, y_train)
            y_predict = est.predict(X)
            incorrect = y_predict != y_train
            estimator_error = np.average(incorrect, weights=sample_weight)
            estimator_weight = (np.log((1 - estimator_error) / estimator_error) +
                                np.log(self.n_classes_ - 1))
            sample_weight *= np.exp(estimator_weight * incorrect)
            sample_weight /= np.sum(sample_weight)
            self.estimators_.append(est)
            self.estimator_errors_[i] = estimator_error
            self.estimator_weights_[i] = estimator_weight
        return self

    def decision_function(self, X):
        pred = np.zeros((X.shape[0], self.n_classes_))
        for i in range(self.n_estimators):
            pred[np.arange(X.shape[0]), self.estimators_[i].predict(X)] += self.estimator_weights_[i]
        pred /= np.sum(self.estimator_weights_)
        if self.n_classes_ == 2:
            return pred[:, 1] - pred[:, 0]
        return pred

    def predict(self, X):
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(int)
        else:
            indices = np.argmax(scores, axis=1)
        return self.classes_[indices]
    


class GradientBoostingClassifier(ClassifierMixin, BaseEstimator):
    def __init__(self, learning_rate=0.1, n_estimators=100, max_depth=3, random_state=0):
        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.random_state = random_state

    def fit(self, X, y):
        self.n_features_ = X.shape[1]
        self.classes_, y = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)
        if self.n_classes_ == 2:
            n_effective_classes = 1
        else:
            n_effective_classes = self.n_classes_
        self.estimators_ = np.empty((self.n_estimators, n_effective_classes), dtype=object)
        raw_predictions = np.zeros((X.shape[0], n_effective_classes))
        for i in range(self.n_estimators):
            raw_predictions_copy = raw_predictions.copy()
            for j in range(n_effective_classes):
                # binary classification
                if n_effective_classes == 1:
                    y_enc = y
                    residual = y_enc - expit(raw_predictions_copy.ravel())
                # multiclass classification
                else:
                    y_enc = (y == j).astype(np.int)
                    residual = y_enc - np.nan_to_num(np.exp(raw_predictions_copy[:, j] 
                                                            - logsumexp(raw_predictions_copy, axis=1)))
                tree = DecisionTreeRegressor(criterion="mse", max_depth=self.max_depth)
                tree.fit(X, residual)
                raw_predictions[:, j] += self.learning_rate * tree.predict(X)
                self.estimators_[i, j] = tree
        self._is_fitted = True
        return self

    def _predict(self, X):
        raw_predictions = np.zeros((X.shape[0], self.estimators_.shape[1]))
        for i in range(self.estimators_.shape[0]):
            for j in range(self.estimators_.shape[1]):
                raw_predictions[:, j] += self.learning_rate * self.estimators_[i, j].predict(X)
        return raw_predictions

    def decision_function(self, X):
        prob = self._predict(X)
        if self.n_classes_ == 2:
            return prob.ravel()
        else:
            return prob

    def predict_proba(self, X):
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            prob = expit(scores)
            prob = np.vstack((1 - prob, prob)).T
        else:
            prob = np.nan_to_num(np.exp(scores - logsumexp(scores, axis=1)[:, np.newaxis]))
        return prob

    def predict(self, X):
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(int)
        else:
            indices = np.argmax(scores, axis=1)
        return self.classes_[indices]

    @property
    def feature_importances_(self):
        all_importances = np.zeros(self.n_features_)
        for i in range(self.estimators_.shape[0]):
            for j in range(self.estimators_.shape[1]):
                all_importances += self.estimators_[i, j].tree_.compute_feature_importances(normalize=False)
        return all_importances / np.sum(all_importances)





import numpy as np
import copy
from base.base import BaseEstimator, TransformerMixin

class SelectFromModel(TransformerMixin, BaseEstimator):
    """Meta-transformer for selecting features based on importance weights.

    This transformer selects features based on the importance weights
    provided by an estimator. Features whose importance exceeds a given
    threshold are retained, while others are discarded.

    Parameters:
    estimator : object
        An estimator with a `fit` method that provides feature importances
        or coefficients.

    Attributes:
    estimator_ : object
        The fitted estimator.
    importances_ : array of shape (n_features,)
        The feature importances or coefficients provided by the estimator.
    threshold_ : float
        The threshold used for feature selection.
    """

    def __init__(self, estimator):
        self.estimator = estimator

    def fit(self, X, y):
        """Fit the SelectFromModel meta-transformer.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,)
            Target values.

        Returns:
        self : SelectFromModel
            The fitted SelectFromModel instance.
        """
        self.estimator_ = copy.deepcopy(self.estimator)
        self.estimator_.fit(X, y)
        
        if hasattr(self.estimator_, "feature_importances_"):
            self.importances_ = self.estimator_.feature_importances_
        elif hasattr(self.estimator_, "coef_"):
            if self.estimator_.coef_.ndim == 1:
                self.importances_ = np.abs(self.estimator_.coef_)
            else:
                self.importances_ = np.linalg.norm(self.estimator_.coef_, ord=1, axis=0)
        
        self.threshold_ = np.mean(self.importances_)
        return self

    def transform(self, X):
        """Transform the input data based on the feature importances.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Data to transform.

        Returns:
        X_transformed : array-like of shape (n_samples, n_selected_features)
            Transformed data with selected features.
        """
        return X[:, self.importances_ >= self.threshold_]



import numpy as np
from numpy.linalg import svd
from base.base import BaseEstimator, TransformerMixin

class PCA(BaseEstimator, TransformerMixin):
    def __init__(self, n_components=2):
        self.n_components = n_components

    def fit(self, X, y=None):
        self.mean_ = np.mean(X, axis=0)
        X_centered = X - self.mean_
        U, S, Vt = svd(X_centered, full_matrices=False)
        self.components_ = Vt[:self.n_components]
        self.explained_variance_ = (S[:self.n_components] ** 2) / (X.shape[0] - 1)
        self.explained_variance_ratio_ = self.explained_variance_ / np.sum(S ** 2 / (X.shape[0] - 1))
        self._is_fitted = True
        return self

    def transform(self, X):
        X_centered = X - self.mean_
        return np.dot(X_centered, self.components_.T)

    def inverse_transform(self, X):
        return np.dot(X, self.components_) + self.mean_


import warnings
warnings.filterwarnings("ignore")

from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.naive_bayes import GaussianNB as skGaussianNB, BernoulliNB as skBernoulliNB, MultinomialNB as skMultinomialNB
from naive_bayes import GaussianNB, BernoulliNB, MultinomialNB

# Load the iris dataset
X, y = load_breast_cancer(return_X_y=True)

# Split the data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=100)

# Initialize the models
my_gnb = GaussianNB()
my_bnb = BernoulliNB()
my_mnb = MultinomialNB()

sk_gnb = skGaussianNB()
sk_bnb = skBernoulliNB()
sk_mnb = skMultinomialNB()

# Train the models
my_gnb.fit(X_train, y_train)
my_bnb.fit(X_train, y_train)
my_mnb.fit(X_train, y_train)

sk_gnb.fit(X_train, y_train)
sk_bnb.fit(X_train, y_train)
sk_mnb.fit(X_train, y_train)

# Calculate accuracy scores
accuracy_my_gnb = accuracy_score(y_test, my_gnb.predict(X_test))
accuracy_my_bnb = accuracy_score(y_test, my_bnb.predict(X_test))
accuracy_my_mnb = accuracy_score(y_test, my_mnb.predict(X_test))

accuracy_sk_gnb = accuracy_score(y_test, sk_gnb.predict(X_test))
accuracy_sk_bnb = accuracy_score(y_test, sk_bnb.predict(X_test))
accuracy_sk_mnb = accuracy_score(y_test, sk_mnb.predict(X_test))

# Print the accuracy scores
print("My Gaussian Naive Bayes Accuracy:", accuracy_my_gnb)
print("My Bernoulli Naive Bayes Accuracy:", accuracy_my_bnb)
print("My Multinomial Naive Bayes Accuracy:", accuracy_my_mnb)

print("Scikit-learn Gaussian Naive Bayes Accuracy:", accuracy_sk_gnb)
print("Scikit-learn Bernoulli Naive Bayes Accuracy:", accuracy_sk_bnb)
print("Scikit-learn Multinomial Naive Bayes Accuracy:", accuracy_sk_mnb)



from um6p_CC_learn.base.base import BaseEstimator, ClassifierMixin, RegressorMixin
import numpy as np


class BaseDecisionTree(BaseEstimator):
    def __init__(self, criterion, max_depth=None, min_samples_split=2, max_features=None):
        """
        Initialize a base decision tree with specified parameters.

        Parameters:
            criterion (str): The function to measure the quality of a split.
            max_depth (int, optional): The maximum depth of the tree.
            min_samples_split (int): The minimum number of samples required to split an internal node.
            max_features (int, float, str or None): The number of features to consider when looking for the best split.
        """
        super().__init__()
        self.max_depth = max_depth if max_depth is not None else np.inf
        self.min_samples_split = min_samples_split
        self.depth = None
        self.criterion = criterion
        self.max_features = max_features if max_features != "auto" else None

    def apply(self, X):
        """
        Apply the decision tree to the input data.

        Parameters:
            X (np.ndarray): Input data.

        Returns:
            np.ndarray: The leaf index for each sample.
        """
        pred = np.zeros(X.shape[0], dtype=int)
        for i in range(X.shape[0]):
            cur_node = 0
            while self.tree_[cur_node].left_child != -1:
                if X[i][self.tree_[cur_node].feature] <= self.tree_[cur_node].threshold:
                    cur_node = self.tree_[cur_node].left_child
                else:
                    cur_node = self.tree_[cur_node].right_child
            pred[i] = cur_node
        return pred
    
    @property
    def feature_importances_(self):
        """
        Calculate the feature importances.

        Returns:
            np.ndarray: Normalized feature importances.
        """
        importances = np.zeros(self.n_features)
        for node in self.tree_:
            if node.left_child != -1:
                left_child = self.tree_[node.left_child]
                right_child = self.tree_[node.right_child]
                importances[node.feature] += (node.n_node * node.impurity
                                              - left_child.n_node * left_child.impurity
                                              - right_child.n_node * right_child.impurity)
        return importances / np.sum(importances)

class TreeNode():
    def __init__(self):
        """
        Initialize a tree node with default values.
        """
        self.left_child = -1
        self.right_child = -1
        self.feature = None
        self.threshold = None
        self.impurity = None
        self.n_node = None
        self.value = None

class DecisionTreeClassifier(ClassifierMixin, BaseDecisionTree):
    
    def __init__(self, criterion='gini', max_depth=None, min_samples_split=2, max_features=None):
        """
        Initialize a decision tree classifier with specified parameters.

        Parameters:
            criterion (str): The function to measure the quality of a split.
            max_depth (int, optional): The maximum depth of the tree.
            min_samples_split (int): The minimum number of samples required to split an internal node.
            max_features (int, float, str or None): The number of features to consider when looking for the best split.
        """
        self._estimator_type = "classifier"
        if criterion not in ['gini', 'entropy']:
            raise ValueError("The criterion for decision tree classifiers should be either 'gini' or 'entropy'.")
         
        super().__init__(criterion, max_depth, min_samples_split, max_features)

    def _gini(self, y_cnt):
        """
        Calculate the Gini impurity.

        Parameters:
            y_cnt (np.ndarray): Array of class counts.

        Returns:
            float: Gini impurity.
        """
        prob = y_cnt / np.sum(y_cnt)
        return 1 - np.sum(np.square(prob))
    
    def _entropy(self, y_cnt):
        """
        Calculate the entropy.

        Parameters:
            y_cnt (np.ndarray): Array of class counts.

        Returns:
            float: Entropy.
        """
        prob = y_cnt / np.sum(y_cnt)
        prob = prob[prob > 0]
        return -np.sum(prob * np.log2(prob))

    def _build_tree(self, X, y, cur_depth, parent, is_left):
        """
        Recursively build the decision tree.

        Parameters:
            X (np.ndarray): Training data.
            y (np.ndarray): Target values.
            cur_depth (int): Current depth of the tree.
            parent (int or None): Parent node index.
            is_left (bool): Whether the current node is a left child.
        """
        if self.criterion == 'gini':
            criterion_func = self._gini
        else:
            criterion_func = self._entropy

        if cur_depth == self.max_depth or np.unique(y).shape[0] == 1 or len(y) < self.min_samples_split:
            cur_node = TreeNode()
            cur_node.n_node = X.shape[0]
            cur_node.value = np.bincount(y, minlength=self.n_classes_)
            cur_node.impurity = criterion_func(cur_node.value)
            cur_id = len(self.tree_)
            self.tree_.append(cur_node)
            if parent is not None:
                if is_left:
                    self.tree_[parent].left_child = cur_id
                else:
                    self.tree_[parent].right_child = cur_id

            if self.depth is None:
                self.depth = cur_depth
            else:
                self.depth = max(cur_depth, self.depth)
            return

        best_improvement = -np.inf
        best_feature = None
        best_threshold = None
        best_left_ind = None
        best_right_ind = None
        y_cnt = np.bincount(y, minlength=self.n_classes_)
        
        # Determine the number of features to consider
        if self.max_features is None:
            features = range(X.shape[1])
        else:
            rng = np.random.RandomState()
            features = rng.choice(X.shape[1], self.max_features, replace=False)
        
        for i in features:
            ind = np.argsort(X[:, i])
            y_cnt_left = np.bincount([], minlength=self.n_classes_)
            y_cnt_right = y_cnt.copy()
            n_left = 0
            n_right = X.shape[0]
            for j in range(ind.shape[0] - 1):
                y_cnt_left[y[ind[j]]] += 1
                y_cnt_right[y[ind[j]]] -= 1
                n_left += 1
                n_right -= 1
                if j + 1 < ind.shape[0] - 1 and np.isclose(X[ind[j], i], X[ind[j + 1], i]):
                    continue
                cur_improvement = -n_left * criterion_func(y_cnt_left) - n_right * criterion_func(y_cnt_right)
                if cur_improvement > best_improvement:
                    best_improvement = cur_improvement
                    best_feature = i
                    best_threshold = X[ind[j], i]
                    best_left_ind = ind[:j + 1]
                    best_right_ind = ind[j + 1:]
        
        cur_node = TreeNode()
        cur_node.feature = best_feature
        cur_node.threshold = best_threshold
        cur_node.n_node = X.shape[0]
        cur_node.value = y_cnt
        cur_node.impurity = criterion_func(y_cnt)
        cur_id = len(self.tree_)
        self.tree_.append(cur_node)
        if parent is not None:
            if is_left:
                self.tree_[parent].left_child = cur_id
            else:
                self.tree_[parent].right_child = cur_id
        if cur_depth < self.max_depth:
            self._build_tree(X[best_left_ind], y[best_left_ind], cur_depth + 1, cur_id, True)
            self._build_tree(X[best_right_ind], y[best_right_ind], cur_depth + 1, cur_id, False)

    def fit(self, X, y):
        """
        Fit the decision tree classifier to the training data.

        Parameters:
            X (np.ndarray): Training data.
            y (np.ndarray): Target values.

        Returns:
            self (DecisionTreeClassifier): Fitted estimator.
        """
        self.n_features = X.shape[1]
        self.classes_, y_train = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)
        self.tree_ = []
        self._build_tree(X, y_train, 0, None, None)

        self._is_fitted = True
        return self

    def predict_proba(self, X):
        """
        Predict class probabilities for X.

        Parameters:
            X (np.ndarray): Input data.

        Returns:
            np.ndarray: Predicted class probabilities.
        """
        self.check_is_fitted()

        pred = self.apply(X)
        prob = np.array([self.tree_[p].value for p in pred])
        return prob / np.sum(prob, axis=1)[:, np.newaxis]

    def predict(self, X):
        """
        Predict class labels for X.

        Parameters:
            X (np.ndarray): Input data.

        Returns:
            np.ndarray: Predicted class labels.
        """
        self.check_is_fitted()

        pred = self.apply(X)
        return np.array([self.classes_[np.argmax(self.tree_[p].value)] for p in pred])

class DecisionTreeRegressor(RegressorMixin, BaseDecisionTree):

    def __init__(self, criterion='mse', max_depth=None, min_samples_split=2, max_features=None):
        """
        Initialize a decision tree regressor with specified parameters.

        Parameters:
            criterion (str): The function to measure the quality of a split.
            max_depth (int, optional): The maximum depth of the tree.
            min_samples_split (int): The minimum number of samples required to split an internal node.
            max_features (int, float, str or None): The number of features to consider when looking for the best split.
        """
        self._estimator_type = "regressor"
        if criterion not in ['mse', 'friedman_mse']:
            raise ValueError("The criterion for decision tree classifiers should be either 'mse' or 'friedman_mse'.")
        
        super().__init__(criterion, max_depth, min_samples_split, max_features)

    def fit(self, X, y):
        """
        Fit the decision tree regressor to the training data.

        Parameters:
            X (np.ndarray): Training data.
            y (np.ndarray): Target values.

        Returns:
            self (DecisionTreeRegressor): Fitted estimator.
        """
        self.n_features = X.shape[1]
        self.tree_ = []
        self._build_tree(X, y, 0, None, None)

        self._is_fitted = True
        return self
    
    def _build_tree(self, X, y, cur_depth, parent, is_left):
        """
        Recursively build the decision tree.

        Parameters:
            X (np.ndarray): Training data.
            y (np.ndarray): Target values.
            cur_depth (int): Current depth of the tree.
            parent (int or None): Parent node index.
            is_left (bool): Whether the current node is a left child.
        """
        if cur_depth == self.max_depth or np.unique(y).shape[0] == 1 or len(y) < self.min_samples_split:
            cur_node = TreeNode()
            cur_node.impurity = np.mean(np.square(y)) - np.square(np.mean(y))
            cur_node.n_node = X.shape[0]
            cur_node.value = np.mean(y)
            cur_id = len(self.tree_)
            self.tree_.append(cur_node)
            if parent is not None:
                if is_left:
                    self.tree_[parent].left_child = cur_id
                else:
                    self.tree_[parent].right_child = cur_id

            if self.depth is None:
                self.depth = cur_depth
            else:
                self.depth = max(cur_depth, self.depth)
                
            return
        
        best_improvement = -np.inf
        best_feature = None
        best_threshold = None
        best_left_ind = None
        best_right_ind = None
        sum_total = np.sum(y)
        
        # Determine the number of features to consider
        if self.max_features is None:
            features = range(X.shape[1])
        else:
            rng = np.random.RandomState()
            features = rng.choice(X.shape[1], self.max_features, replace=False)
        
        for i in features:
            sum_left = 0
            sum_right = sum_total
            n_left = 0
            n_right = X.shape[0]
            ind = np.argsort(X[:, i])
            for j in range(ind.shape[0] - 1):
                sum_left += y[ind[j]]
                sum_right -= y[ind[j]]
                n_left += 1
                n_right -= 1
                if j + 1 < ind.shape[0] - 1 and np.isclose(X[ind[j], i], X[ind[j + 1], i]):
                    continue
                cur_improvement = self._calc_improvement(sum_left, n_left, sum_right, n_right)
                if cur_improvement > best_improvement:
                    best_improvement = cur_improvement
                    best_feature = i
                    best_threshold = X[ind[j], i]
                    best_left_ind = ind[:j + 1]
                    best_right_ind = ind[j + 1:]
        
        cur_node = TreeNode()
        cur_node.feature = best_feature
        cur_node.threshold = best_threshold
        cur_node.impurity = np.mean(np.square(y)) - np.square(np.mean(y))
        cur_node.n_node = X.shape[0]
        cur_node.value = np.mean(y)
        cur_id = len(self.tree_)
        self.tree_.append(cur_node)
        if parent is not None:
            if is_left:
                self.tree_[parent].left_child = cur_id
            else:
                self.tree_[parent].right_child = cur_id
        if cur_depth < self.max_depth:
            self._build_tree(X[best_left_ind], y[best_left_ind], cur_depth + 1, cur_id, True)
            self._build_tree(X[best_right_ind], y[best_right_ind], cur_depth + 1, cur_id, False)

    def _calc_improvement(self, sum_left, n_left, sum_right, n_right):
        """
        Calculate the improvement of a split.

        Parameters:
            sum_left (float): Sum of the left split target values.
            n_left (int): Number of samples in the left split.
            sum_right (float): Sum of the right split target values.
            n_right (int): Number of samples in the right split.

        Returns:
            float: Improvement value.
        """
        if self.criterion == 'mse':
            return sum_left * sum_left / n_left + sum_right * sum_right / n_right
        elif self.criterion == 'friedman_mse':
            return n_left * n_right * np.square(sum_left / n_left - sum_right / n_right)
    
    def predict(self, X):
        """
        Predict target values for X.

        Parameters:
            X (np.ndarray): Input data.

        Returns:
            np.ndarray: Predicted target values.
        """
        self.check_is_fitted()

        pred = self.apply(X)
        return np.array([self.tree_[p].value for p in pred])


import numpy as np
import copy
from base.base import TransformerMixin, BaseEstimator

class RFE(TransformerMixin, BaseEstimator):
    """Feature ranking with recursive feature elimination (RFE).

    This transformer recursively removes features, building a model on the
    remaining features at each step until the desired number of features is
    reached.

    Parameters:
    estimator : object
        An estimator with a `fit` method that provides feature importances
        or coefficients.

    Attributes:
    estimator_ : object
        The fitted estimator.
    support_ : ndarray of shape (n_features,)
        The mask of selected features.
    ranking_ : ndarray of shape (n_features,)
        The feature ranking, with the selected features assigned rank 1.
    """

    def __init__(self, estimator):
        self.estimator = estimator

    def fit(self, X, y):
        """Fit the RFE meta-transformer.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,)
            Target values.

        Returns:
        self : RFE
            The fitted RFE instance.
        """
        n_features_to_select = X.shape[1] // 2
        support = np.ones(X.shape[1], dtype=np.bool)
        ranking = np.ones(X.shape[1], dtype=np.int)
        
        while np.sum(support) > n_features_to_select:
            est = copy.deepcopy(self.estimator)
            est.fit(X[:, support], y)
            
            if hasattr(est, "feature_importances_"):
                importances = est.feature_importances_
            elif hasattr(est, "coef_"):
                if est.coef_.ndim == 1:
                    importances = np.abs(est.coef_)
                else:
                    importances = np.linalg.norm(est.coef_, ord=1, axis=0)
            
            cur_feature = np.arange(X.shape[1])[support][np.argmin(importances)]
            support[cur_feature] = False
            ranking[~support] += 1
        
        self.support_ = support
        self.ranking_ = ranking
        self.estimator_ = copy.deepcopy(self.estimator)
        self.estimator_.fit(X[:, support], y)
        return self

    def transform(self, X):
        """Transform the input data based on selected features.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Data to transform.

        Returns:
        X_transformed : array-like of shape (n_samples, n_selected_features)
            Transformed data with selected features.
        """
        return X[:, self.support_]


import copy

def clone(estimator, *, safe=True):
    """Construct a new unfitted estimator with the same parameters.

    Clone does a deep copy of the model in an estimator
    without actually copying attached data. It returns a new estimator
    with the same parameters that has not been fitted on any data.

    .. versionchanged:: 1.3
        Delegates to `estimator.__sklearn_clone__` if the method exists.

    Parameters
    ----------
    estimator : {list, tuple, set} of estimator instance or a single \
            estimator instance
        The estimator or group of estimators to be cloned.
    safe : bool, default=True
        If safe is False, clone will fall back to a deep copy on objects
        that are not estimators. Ignored if `estimator.__sklearn_clone__`
        exists.

    Returns
    -------
    estimator : object
        The deep copy of the input, an estimator if input is an estimator.

    Notes
    -----
    If the estimator's `random_state` parameter is an integer (or if the
    estimator doesn't have a `random_state` parameter), an *exact clone* is
    returned: the clone and the original estimator will give the exact same
    results. Otherwise, *statistical clone* is returned: the clone might
    return different results from the original estimator. More details can be
    found in :ref:`randomness`.

    Examples
    --------
    >>> from sklearn.base import clone
    >>> from sklearn.linear_model import LogisticRegression
    >>> X = [[-1, 0], [0, 1], [0, -1], [1, 0]]
    >>> y = [0, 0, 1, 1]
    >>> classifier = LogisticRegression().fit(X, y)
    >>> cloned_classifier = clone(classifier)
    >>> hasattr(classifier, "classes_")
    True
    >>> hasattr(cloned_classifier, "classes_")
    False
    >>> classifier is cloned_classifier
    False
    """
    if safe:
        return copy.deepcopy(estimator)
    else:
        return copy.copy(estimator)



import numpy as np
from scipy.optimize import minimize
from base.base import BaseEstimator, ClassifierMixin

class LogisticRegression(ClassifierMixin, BaseEstimator):
    """
    Logistic Regression classifier using maximum likelihood estimation.

    Attributes:
        coef_ (np.ndarray): Coefficients of the logistic regression model.
        intercept_ (float): Intercept of the logistic regression model.
    """

    def sigmoid(self, z):
        """
        Compute the sigmoid function.

        Parameters:
            z (np.ndarray): Input array.

        Returns:
            np.ndarray: Sigmoid of the input array.
        """
        return 1 / (1 + np.exp(-z))

    def log_likelihood(self, coef, X, y):
        """
        Compute the log-likelihood of the logistic regression model.

        Parameters:
            coef (np.ndarray): Coefficients of the model including intercept.
            X (np.ndarray): Training data.
            y (np.ndarray): Target values.

        Returns:
            float: Log-likelihood value.
        """
        z = np.dot(X, coef)
        ll = np.sum(y * z - np.log(1 + np.exp(z)))
        return -ll  # Negative log-likelihood for minimization

    def fit(self, X, y):
        """
        Fit the logistic regression model using the training data.

        Parameters:
            X (np.ndarray): Training data of shape (n_samples, n_features).
            y (np.ndarray): Target values of shape (n_samples,).

        Returns:
            self (LogisticRegression): Returns the instance itself.
        """
        # Add a column of ones to X to account for the intercept
        X_with_intercept = np.hstack((np.ones((X.shape[0], 1)), X))
        
        # Initial guess for the coefficients
        initial_coef = np.zeros(X_with_intercept.shape[1])
        
        # Minimize the negative log-likelihood
        result = minimize(self.log_likelihood, initial_coef, args=(X_with_intercept, y), method='BFGS')
        
        # Store the coefficients and intercept
        self.intercept_ = result.x[0]
        self.coef_ = result.x[1:]

        self._is_fitted = True
        return self

    def predict_proba(self, X):
        """
        Predict probability estimates for the input data.

        Parameters:
            X (np.ndarray): Input data of shape (n_samples, n_features).

        Returns:
            np.ndarray: Predicted probabilities of shape (n_samples, 2).
        """
        self.check_is_fitted()

        # Add a column of ones to X to account for the intercept
        X_with_intercept = np.hstack((np.ones((X.shape[0], 1)), X))
        
        # Compute the probability estimates
        prob = self.sigmoid(np.dot(X_with_intercept, np.hstack(([self.intercept_], self.coef_))))
        
        return np.vstack((1 - prob, prob)).T

    def predict(self, X):
        """
        Predict class labels for the input data.

        Parameters:
            X (np.ndarray): Input data of shape (n_samples, n_features).

        Returns:
            np.ndarray: Predicted class labels of shape (n_samples,).
        """
        self.check_is_fitted()

        # Predict probabilities
        prob = self.predict_proba(X)[:, 1]
        
        # Predict class labels based on probability threshold of 0.5
        return (prob >= 0.5).astype(int)
    def get_params(self, deep=True):
        """
        Get parameters for this estimator.

        Parameters:
            deep (bool, optional): If True, will return the parameters for this estimator and contained subobjects that are estimators.

        Returns:
            dict: Parameter names mapped to their values.
        """
        return {}


import numpy as np
from base.base import BaseEstimator
from utils.__clone import clone
from model_selection.__split import KFold, StratifiedKFold
from itertools import product

class GridSearchCVBase(BaseEstimator):
    def __init__(self, estimator, param_grid):
        self.estimator = estimator
        self.param_grid = param_grid

    def generate_grid(self):
        items = sorted(self.param_grid.items())
        keys, values = zip(*items)
        for v in product(*values):
            params = dict(zip(keys, v))
            yield params

    def fit(self, X, y):
        cv = self.get_cv(y)
        train_scores, test_scores  = [], []
        params = []
        for i, cur_param in enumerate(self.generate_grid()):
            cur_train_score, cur_test_score = [], []
            for j, (train, test) in enumerate(cv.split(X, y)):
                est = clone(self.estimator)
                est.set_params(**cur_param)
                est.fit(X[train], y[train])
                cur_train_score.append(est.score(X[train], y[train]))
                cur_test_score.append(est.score(X[test], y[test]))
            params.append(cur_param)
            train_scores.append(cur_train_score)
            test_scores.append(cur_test_score)
        train_scores = np.array(train_scores)
        test_scores = np.array(test_scores)
        cv_results = {}
        for i in range(cv.n_splits):
            cv_results["split" + str(i) + "_train_score"] = train_scores[:, i]
            cv_results["split" + str(i) + "_test_score"] = test_scores[:, i]
        cv_results["mean_train_score"] = np.mean(train_scores, axis=1)
        cv_results["std_train_score"] = np.std(train_scores, axis=1)
        cv_results["mean_test_score"] = np.mean(test_scores, axis=1)
        cv_results["std_test_score"] = np.std(test_scores, axis=1)
        cv_results['params'] = params
        self.cv_results_ = cv_results
        self.best_params_ = cv_results['params'][np.argmax(cv_results['mean_test_score'])]
        self.best_estimator_ = clone(self.estimator)
        self.best_estimator_.set_params(**self.best_params_)
        self.best_estimator_.fit(X, y)
        return self

    def get_cv(self, y):
        raise NotImplementedError

    def decision_function(self, X):
        return self.best_estimator_.decision_function(X)

    def predict(self, X):
        return self.best_estimator_.predict(X)


class GridSearchCVClassifier(GridSearchCVBase):
    def get_cv(self, y):
        return StratifiedKFold()


class GridSearchCVRegressor(GridSearchCVBase):
    def get_cv(self, y):
        return KFold()


import numpy as np
import time
from sklearn.datasets import load_breast_cancer
from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import KNeighborsClassifier as skKNeighborsClassifier
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from neighbors.KNeighborsClassifier import KNeighborsClassifier

X, y = load_breast_cancer(return_X_y=True)

X = StandardScaler().fit_transform(X)

start_time = time.time()
clf1 = KNeighborsClassifier().fit(X, y)
end_time = time.time()
my_fit_time = end_time - start_time

start_time = time.time()
clf2 = skKNeighborsClassifier().fit(X, y)
end_time = time.time()
sk_fit_time = end_time - start_time

prob1 = clf1.predict_proba(X)

# Predict probabilities using scikit-learn's KNeighborsClassifier
prob2 = clf2.predict_proba(X)

# Check if probabilities are almost equal
assert np.allclose(prob1, prob2), "Probability predictions do not match."

# Predict classes using my KNeighborsClassifier
pred1 = clf1.predict(X)

# Predict classes using scikit-learn's KNeighborsClassifier
pred2 = clf2.predict(X)

# Check if class predictions are equal
assert np.array_equal(pred1, pred2), "Class predictions do not match."

# Calculate evaluation metrics
accuracy_my = accuracy_score(y, pred1)
precision_my = precision_score(y, pred1)
recall_my = recall_score(y, pred1)
f1_my = f1_score(y, pred1)

accuracy_sk = accuracy_score(y, pred2)
precision_sk = precision_score(y, pred2)
recall_sk = recall_score(y, pred2)
f1_sk = f1_score(y, pred2)

# Printing the results
print("My KNeighborsClassifier")
print("Accuracy:", accuracy_my)
print("Precision:", precision_my)
print("Recall:", recall_my)
print("F1-score:", f1_my)
print("Time taken to fit:", my_fit_time)

print("\nScikit-learn's KNeighborsClassifier")
print("Accuracy:", accuracy_sk)
print("Precision:", precision_sk)
print("Recall:", recall_sk)
print("F1-score:", f1_sk)
print("Time taken to fit:", sk_fit_time)



from utils.__clone import clone

from model_selection.__validation import cross_val_predict

from base.base import RegressorMixin, BaseEstimator

import numpy as np

class StackingClassifier(RegressorMixin, BaseEstimator):
    def __init__(self, estimators, final_estimator):
        self.estimators = estimators
        self.final_estimator = final_estimator

    def fit(self, X, y):
        self.estimators_ = []
        for est in self.estimators:
            self.estimators_.append(clone(est).fit(X, y))
        predictions = []
        for est in self.estimators:
            cur_prediction = cross_val_predict(est, X, y, method="predict_proba")
            if cur_prediction.shape[1] == 2:
                predictions.append(cur_prediction[:, [1]])
            else:
                predictions.append(cur_prediction)
        X_meta = np.hstack(predictions)
        self.final_estimator_ = clone(self.final_estimator)
        self.final_estimator_.fit(X_meta, y)
        self._is_fitted = True
        return self

    def transform(self, X):
        predictions = []
        for est in self.estimators_:
            cur_prediction = est.predict_proba(X)
            if cur_prediction.shape[1] == 2:
                predictions.append(cur_prediction[:, [1]])
            else:
                predictions.append(cur_prediction)
        return np.hstack(predictions)

    def predict(self, X):
        return self.final_estimator_.predict(self.transform(X))

    def predict_proba(self, X):
        return self.final_estimator_.predict_proba(self.transform(X))
    

class StackingRegressor(RegressorMixin, BaseEstimator):
    def __init__(self, estimators, final_estimator):
        self.estimators = estimators
        self.final_estimator = final_estimator

    def fit(self, X, y):
        self.estimators_ = [clone(est).fit(X, y) for est in self.estimators]
        predictions = [
            cross_val_predict(est, X, y, method='predict').reshape(-1, 1)
            for est in self.estimators_
        ]
        X_meta = np.hstack(predictions)
        self.final_estimator_ = clone(self.final_estimator).fit(X_meta, y)
        self._is_fitted = True
        return self

    def transform(self, X):
        predictions = [
            est.predict(X).reshape(-1, 1) for est in self.estimators_
        ]
        return np.hstack(predictions)

    def predict(self, X):
        return self.final_estimator_.predict(self.transform(X))


import numpy as np
import copy
from base.base import BaseEstimator, ClassifierMixin, RegressorMixin
from tree.decision_tree import DecisionTreeClassifier, DecisionTreeRegressor

class BaggingClassifier(ClassifierMixin, BaseEstimator):
    def __init__(self, base_estimator=None, n_estimators=10, max_samples=1.0, bootstrap=True, random_state=None):
        self.base_estimator = base_estimator if base_estimator is not None else DecisionTreeClassifier()
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.bootstrap = bootstrap
        self.random_state = random_state
        self.estimators_ = []

    def fit(self, X, y):
        np.random.seed(self.random_state)
        self.estimators_ = []

        for _ in range(self.n_estimators):
            estimator = copy.deepcopy(self.base_estimator)
            n_samples = int(self.max_samples * X.shape[0])

            if self.bootstrap:
                indices = np.random.choice(X.shape[0], n_samples, replace=True)
            else:
                indices = np.random.choice(X.shape[0], n_samples, replace=False)

            estimator.fit(X[indices], y[indices])
            self.estimators_.append(estimator)

        return self

    def predict(self, X):
        predictions = np.array([estimator.predict(X) for estimator in self.estimators_])
        return np.apply_along_axis(lambda x: np.bincount(x).argmax(), axis=0, arr=predictions)


class BaggingRegressor(RegressorMixin, BaseEstimator):
    def __init__(self, base_estimator=None, n_estimators=10, max_samples=1.0, bootstrap=True, random_state=None):
        self.base_estimator = base_estimator if base_estimator is not None else DecisionTreeRegressor()
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.bootstrap = bootstrap
        self.random_state = random_state
        self.estimators_ = []

    def fit(self, X, y):
        np.random.seed(self.random_state)
        self.estimators_ = []

        for _ in range(self.n_estimators):
            estimator = copy.deepcopy(self.base_estimator)
            n_samples = int(self.max_samples * X.shape[0])

            if self.bootstrap:
                indices = np.random.choice(X.shape[0], n_samples, replace=True)
            else:
                indices = np.random.choice(X.shape[0], n_samples, replace=False)

            estimator.fit(X[indices], y[indices])
            self.estimators_.append(estimator)

        return self

    def predict(self, X):
        predictions = np.array([estimator.predict(X) for estimator in self.estimators_])
        return np.mean(predictions, axis=0)



import numpy as np
from scipy.sparse import coo_matrix


def accuracy_score(y_true, y_pred):
    return np.mean(y_true == y_pred)

def fbeta_score(y_true, y_pred, beta, average="binary"):

    n_labels = len(set(y_true) | set(y_pred))
    true_sum = np.bincount(y_true, minlength=n_labels)
    pred_sum = np.bincount(y_pred, minlength=n_labels)
    tp = np.bincount(y_true[y_true == y_pred], minlength=n_labels)
    
    if average == "binary":
        tp = np.array([tp[1]])
        true_sum = np.array([true_sum[1]])
        pred_sum = np.array([pred_sum[1]])
    elif average == "micro":
        tp = np.array([np.sum(tp)])
        true_sum = np.array([np.sum(true_sum)])
        pred_sum = np.array([np.sum(pred_sum)])
        
    precision = np.zeros(len(pred_sum))
    mask = pred_sum != 0
    precision[mask] = tp[mask] / pred_sum[mask]
    
    recall = np.zeros(len(true_sum))
    mask = true_sum != 0
    recall[mask] = tp[mask] / true_sum[mask]
    
    denom = (beta ** 2) * precision + recall
    denom[denom == 0.] = 1
    
    fscore = (1 + np.square(beta)) * precision * recall / denom
    
    if average == "weighted":
        fscore = np.average(fscore, weights=true_sum)
    elif average is not None:
        fscore = np.mean(fscore)
    
    return fscore


def f1_score(y_true, y_pred, average="binary"):
    return fbeta_score(y_true, y_pred, 1.0, average)


def recall_score(y_true, y_pred, average):
    n_labels = len(set(y_true) | set(y_pred))
    true_sum = np.bincount(y_true, minlength=n_labels)
    tp = np.bincount(y_true[y_true == y_pred], minlength=n_labels)
    
    if average == "binary":
        tp = np.array([tp[1]])
        true_sum = np.array([true_sum[1]])
    elif average == "micro":
        tp = np.array([np.sum(tp)])
        true_sum = np.array([np.sum(true_sum)])
    
    recall = np.zeros(len(true_sum))
    mask = true_sum != 0
    recall[mask] = tp[mask] / true_sum[mask]
    
    if average == "weighted":
        recall = np.average(recall, weights=true_sum)
    elif average is not None:
        recall = np.mean(recall)
    
    return recall



def precision_score(y_true, y_pred, average = "binary"):

    n_labels = len(set(y_true) | set(y_pred))
    true_sum = np.bincount(y_true, minlength=n_labels)
    pred_sum = np.bincount(y_pred, minlength=n_labels)
    tp = np.bincount(y_true[y_true == y_pred], minlength=n_labels)

    if average == "binary":
        tp = np.array([tp[1]])
        pred_sum = np.array([pred_sum[1]])
    elif average == "micro":
        tp = np.array([np.sum(tp)])
        pred_sum = np.array([np.sum(pred_sum)])
    
    precision = np.zeros(len(pred_sum))
    mask = pred_sum != 0
    precision[mask] = tp[mask] / pred_sum[mask]

    if average == "weighted":
        precision = np.average(precision, weights=true_sum)
    elif average is not None:
        precision = np.mean(precision)
    return precision


def confusion_matrix(y_true, y_pred):
    labels = np.unique(np.concatenate([y_true, y_pred]))
    n_labels = len(labels)
    label_to_ind = {label: i for i, label in enumerate(labels)}
    
    true_inds = np.array([label_to_ind[label] for label in y_true])
    pred_inds = np.array([label_to_ind[label] for label in y_pred])
    
    cm = np.bincount(n_labels * true_inds + pred_inds, minlength=np.square(n_labels)).reshape(n_labels, n_labels)
    
    return cm





import numpy as np
from base.base import ClassifierMixin, RegressorMixin, BaseEstimator
from tree.decision_tree import DecisionTreeClassifier, DecisionTreeRegressor

class RandomForestClassifier(ClassifierMixin, BaseEstimator):
    def __init__(self, n_estimators=100, criterion='gini', max_depth=None, min_samples_split=2, max_features='auto', bootstrap=True, random_state=None):
        """
        Initialize a random forest classifier with specified parameters.

        Parameters:
            n_estimators (int): The number of trees in the forest.
            criterion (str): The function to measure the quality of a split.
            max_depth (int, optional): The maximum depth of the tree.
            min_samples_split (int): The minimum number of samples required to split an internal node.
            max_features (int, float, str or None): The number of features to consider when looking for the best split.
            bootstrap (bool): Whether bootstrap samples are used when building trees.
            random_state (int, optional): The random seed.
        """
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.random_state = random_state
        self.estimators_ = []
        self._is_fitted = False
        self._estimator_type = "classifier"

    def fit(self, X, y):
        """
        Fit the random forest classifier to the training data.

        Parameters:
            X (np.ndarray): Training data.
            y (np.ndarray): Target values.

        Returns:
            self (RandomForestClassifier): Fitted estimator.
        """
        if self.random_state is not None:
            np.random.seed(self.random_state)
        
        self.n_samples_, self.n_features_ = X.shape
        self.classes_, y_train = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)
        
        for _ in range(self.n_estimators):
            tree = DecisionTreeClassifier(
                criterion=self.criterion,
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                max_features=self.max_features
            )
            if self.bootstrap:
                indices = np.random.choice(self.n_samples_, self.n_samples_, replace=True)
            else:
                indices = np.arange(self.n_samples_)
            tree.fit(X[indices], y_train[indices])
            self.estimators_.append(tree)

        self._is_fitted = True
        return self

    def predict_proba(self, X):
        """
        Predict class probabilities for X.

        Parameters:
            X (np.ndarray): Input data.

        Returns:
            np.ndarray: Predicted class probabilities.
        """
        self.check_is_fitted()

        all_probas = np.zeros((X.shape[0], self.n_classes_))
        for tree in self.estimators_:
            all_probas += tree.predict_proba(X)
        all_probas /= self.n_estimators
        return all_probas

    def predict(self, X):
        """
        Predict class labels for X.

        Parameters:
            X (np.ndarray): Input data.

        Returns:
            np.ndarray: Predicted class labels.
        """
        self.check_is_fitted()

        proba = self.predict_proba(X)
        return self.classes_[np.argmax(proba, axis=1)]

    @property
    def feature_importances_(self):
        """
        Feature importances for the random forest.

        Returns:
            np.ndarray: Normalized feature importances.
        """
        self.check_is_fitted()
        all_importances = np.zeros(self.n_features_)
        
        for est in self.estimators_:
            all_importances += est.feature_importances_()
        
        all_importances /= self.n_estimators
        return all_importances / np.sum(all_importances)
    

class RandomForestRegressor(RegressorMixin, BaseEstimator):
    def __init__(self, n_estimators=100, criterion='mse', max_depth=None, min_samples_split=2, max_features='auto', bootstrap=True, random_state=None):
        """
        Initialize a random forest regressor with specified parameters.

        Parameters:
            n_estimators (int): The number of trees in the forest.
            criterion (str): The function to measure the quality of a split.
            max_depth (int, optional): The maximum depth of the tree.
            min_samples_split (int): The minimum number of samples required to split an internal node.
            max_features (int, float, str or None): The number of features to consider when looking for the best split.
            bootstrap (bool): Whether bootstrap samples are used when building trees.
            random_state (int, optional): The random seed.
        """
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.random_state = random_state
        self.estimators_ = []
        self._is_fitted = False
        self._estimator_type = "regressor"

    def fit(self, X, y):
        """
        Fit the random forest regressor to the training data.

        Parameters:
            X (np.ndarray): Training data.
            y (np.ndarray): Target values.

        Returns:
            self (RandomForestRegressor): Fitted estimator.
        """
        if self.random_state is not None:
            np.random.seed(self.random_state)

        self.n_samples_, self.n_features_ = X.shape
        
        for _ in range(self.n_estimators):
            tree = DecisionTreeRegressor(
                criterion=self.criterion,
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                max_features=self.max_features
            )
            if self.bootstrap:
                indices = np.random.choice(self.n_samples_, self.n_samples_, replace=True)
            else:
                indices = np.arange(self.n_samples_)
            tree.fit(X[indices], y[indices])
            self.estimators_.append(tree)

        self._is_fitted = True
        return self

    def predict(self, X):
        """
        Predict regression target for X.

        Parameters:
            X (np.ndarray): Input data.

        Returns:
            np.ndarray: Predicted target values.
        """
        self.check_is_fitted()

        all_preds = np.zeros(X.shape[0])
        for tree in self.estimators_:
            all_preds += tree.predict(X)
        return all_preds / self.n_estimators
    
    @property
    def feature_importances_(self):
        """
        Feature importances for the random forest.

        Returns:
            np.ndarray: Normalized feature importances.
        """
        self.check_is_fitted()
        all_importances = np.zeros(self.n_features_)
        
        for est in self.estimators_:
            all_importances += est.feature_importances_()
        
        all_importances /= self.n_estimators
        return all_importances / np.sum(all_importances)


import warnings
warnings.filterwarnings("ignore")

from sklearn.datasets import load_breast_cancer, load_iris
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.tree import DecisionTreeClassifier as skDecisionTreeClassifier
from tree.decision_tree import DecisionTreeClassifier

X, y = load_breast_cancer(return_X_y=True)

# Splitting the data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=100)

criterion = 'gini'
max_depth = None

# Training my decision tree classifier
my_clf = DecisionTreeClassifier(criterion=criterion, max_depth=max_depth).fit(X_train, y_train)

# Training sklearn's decision tree classifier
sk_clf = skDecisionTreeClassifier(random_state=0, criterion=criterion, max_depth=max_depth).fit(X_train, y_train)

# Calculating accuracy scores
accuracy_score_your_model = accuracy_score(y_test, my_clf.predict(X_test))
accuracy_score_sklearn_model = accuracy_score(y_test, sk_clf.predict(X_test))

# Printing the accuracy scores
print("Accuracy score of my decision tree classifier:", accuracy_score_your_model)
print("Depth of my decision tree classifier:", my_clf.depth)

print("Accuracy score of sklearn's decision tree classifier:", accuracy_score_sklearn_model)
print("Depth of sklearn's decision tree classifier:", sk_clf.get_depth())

print("--------------------------------------------------------------------------------")

from sklearn.datasets import fetch_california_housing
from sklearn.metrics import mean_squared_error
from sklearn.tree import DecisionTreeRegressor as skDecisionTreeRegressor
from tree.decision_tree import DecisionTreeRegressor

# Load the Boston housing dataset
X, y = fetch_california_housing(return_X_y=True)

# Splitting the data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=100)

# Setting parameters
criterion = 'mse'
max_depth = None

# Training your decision tree regressor
my_reg = DecisionTreeRegressor(criterion=criterion, max_depth=max_depth).fit(X_train, y_train)

# Training scikit-learn's decision tree regressor
sk_reg = skDecisionTreeRegressor(random_state=0, criterion=criterion, max_depth=max_depth).fit(X_train, y_train)

# Calculating mean squared error
mse_your_model = mean_squared_error(y_test, my_reg.predict(X_test))
mse_sklearn_model = mean_squared_error(y_test, sk_reg.predict(X_test))

# Printing the mean squared error
print("Mean Squared Error of my decision tree regressor:", mse_your_model)
print("Depth of my decision tree regressor:", my_reg.depth)

print("Mean Squared Error of scikit-learn's decision tree regressor:", mse_sklearn_model)
print("Depth of scikit-learn's decision tree regressor:", sk_reg.get_depth())



import numpy as np
from base.base import TransformerMixin, BaseEstimator

class VarianceThreshold(TransformerMixin, BaseEstimator):
    """Feature selector that removes low-variance features.

    This transformer removes features whose variance does not meet a certain
    threshold. By default, it removes features with zero variance.

    Parameters:
    threshold : float, default=0
        The variance threshold below which features will be removed.

    Attributes:
    variances_ : array of shape (n_features,)
        The variance of each feature in the training data.
    """

    def __init__(self, threshold=0):
        self.threshold = threshold

    def fit(self, X, y=None):
        """Compute the variance of each feature in X.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,), default=None
            Target values (ignored).

        Returns:
        self : VarianceThreshold
            The fitted VarianceThreshold instance.
        """
        self.variances_ = np.var(X, axis=0)
        return self

    def transform(self, X):
        """Remove features with variance below the threshold.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Data to transform.

        Returns:
        X_transformed : array-like of shape (n_samples, n_features)
            Transformed data with low-variance features removed.
        """
        return X[:, self.variances_ > self.threshold]


import numpy as np
from base.base import BaseEstimator, TransformerMixin

class OrdinalEncoder(TransformerMixin, BaseEstimator):
    def fit(self, X):
        """
        Fit OrdinalEncoder to X.

        Parameters:
        X (np.ndarray): The input array of shape (n_samples, n_features).
        """
        self.categories_ = [np.unique(column) for column in X.T]
        return self
    
    def transform(self, X):
        """
        Transform X using ordinal encoding.

        Parameters:
        X (np.ndarray): The input array of shape (n_samples, n_features).
        
        Returns:
        np.ndarray: Transformed array of shape (n_samples, n_features).
        """
        encoded_columns = []
        for i, column in enumerate(X.T):
            category_indices = {category: idx for idx, category in enumerate(self.categories_[i])}
            encoded_column = np.array([category_indices[value] for value in column])
            encoded_columns.append(encoded_column[:, np.newaxis])
        
        return np.hstack(encoded_columns)
    
    def fit_transform(self, X):
        """
        Fit to data, then transform it.

        Parameters:
        X (np.ndarray): The input array of shape (n_samples, n_features).
        
        Returns:
        np.ndarray: Transformed array of shape (n_samples, n_features).
        """
        return self.fit(X).transform(X)

# Example usage:
if __name__ == "__main__":
    X = np.array([['cat', 'A'], ['dog', 'B'], ['cat', 'A'], ['bird', 'B']])
    ordinal_encoder = OrdinalEncoder()
    X_encoded = ordinal_encoder.fit_transform(X)
    print(X_encoded)



import numpy as np
from base.base import TransformerMixin, BaseEstimator

class MissingIndicator(TransformerMixin, BaseEstimator):
    def __init__(self, missing_values=np.nan, features='missing-only', sparse=False):
        self.missing_values = missing_values
        self.features = features
        self.sparse = sparse

    def fit(self, X, y=None):
        X = np.asarray(X)
        if self.features not in ['missing-only', 'all']:
            raise ValueError("Invalid value for 'features': {}. Allowed values are 'missing-only' or 'all'.".format(self.features))

        if self.missing_values is np.nan:
            mask = np.isnan(X)
        else:
            mask = (X == self.missing_values)

        if self.features == 'missing-only':
            n_missing = np.sum(mask, axis=0)
            self.features_ = np.flatnonzero(n_missing)
        else:
            self.features_ = np.arange(X.shape[1])

        return self

    def transform(self, X):
        X = np.asarray(X)
        if self.missing_values is np.nan:
            mask = np.isnan(X)
        else:
            mask = (X == self.missing_values)

        if self.features == 'missing-only':
            mask = mask[:, self.features_]
        elif self.features == 'all':
            mask = mask

        if self.sparse:
            from scipy.sparse import csr_matrix
            mask = csr_matrix(mask)

        return mask

    def fit_transform(self, X, y=None):
        return self.fit(X, y).transform(X)





import numpy as np


def mean_absolute_error(y_true, y_pred):
    return np.mean(np.abs(y_true - y_pred))


def mean_squared_error(y_true, y_pred):
    return np.mean(np.square((y_true - y_pred)))

def r2_score(y_true, y_pred):
    numerator = np.sum(np.square((y_true - y_pred)))
    denominator = np.sum(np.square((y_true - np.mean(y_true))))
    return 1 - numerator / denominator



import numpy as np
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.neural_network import MLPClassifier
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from neural_network.multilayer_perceptron import NeuralNetwork
import time

# Load data
data = load_iris()
X = data.data
y = data.target

# Split data
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=42)

# Create and train custom neural network
nn = NeuralNetwork(layers=[X_train.shape[1], 10, len(np.unique(y))], learning_rate=0.01, epochs=10000)
start_time = time.time()
nn.fit(X_train, y_train)
nn_fit_time = time.time() - start_time

# Predict with custom neural network
nn_predictions = nn.predict(X_test)

# Metrics for custom neural network
nn_accuracy = accuracy_score(y_test, nn_predictions)
nn_precision = precision_score(y_test, nn_predictions, average='weighted')
nn_recall = recall_score(y_test, nn_predictions, average='weighted')
nn_f1 = f1_score(y_test, nn_predictions, average='weighted')

print(f"Custom Neural Network - Fit Time: {nn_fit_time:.4f}s, Accuracy: {nn_accuracy:.4f}, Precision: {nn_precision:.4f}, Recall: {nn_recall:.4f}, F1 Score: {nn_f1:.4f}")

# Create and train scikit-learn neural network
sk_nn = MLPClassifier(hidden_layer_sizes=(10,), max_iter=10000, learning_rate_init=0.01, random_state=42)
start_time = time.time()
sk_nn.fit(X_train, y_train)
sk_nn_fit_time = time.time() - start_time

# Predict with scikit-learn neural network
sk_nn_predictions = sk_nn.predict(X_test)

# Metrics for scikit-learn neural network
sk_nn_accuracy = accuracy_score(y_test, sk_nn_predictions)
sk_nn_precision = precision_score(y_test, sk_nn_predictions, average='weighted')
sk_nn_recall = recall_score(y_test, sk_nn_predictions, average='weighted')
sk_nn_f1 = f1_score(y_test, sk_nn_predictions, average='weighted')

print(f"Scikit-learn Neural Network - Fit Time: {sk_nn_fit_time:.4f}s, Accuracy: {sk_nn_accuracy:.4f}, Precision: {sk_nn_precision:.4f}, Recall: {sk_nn_recall:.4f}, F1 Score: {sk_nn_f1:.4f}")



import numpy as np
from base.base import BaseEstimator, ClassifierMixin

class NeuralNetwork(ClassifierMixin, BaseEstimator):
    def __init__(self, layers, learning_rate=0.01, epochs=1000):
        """
        Initialize the neural network.

        Parameters:
        layers (list): List of integers representing the number of nodes in each layer.
        learning_rate (float): Learning rate for gradient descent.
        epochs (int): Number of iterations for training.
        """
        self.layers = layers
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.weights = []
        self.biases = []
        
        # Initialize weights and biases
        for i in range(len(layers) - 1):
            weight_matrix = np.random.randn(layers[i], layers[i + 1]) * 0.01
            bias_vector = np.zeros((1, layers[i + 1]))
            self.weights.append(weight_matrix)
            self.biases.append(bias_vector)

    def softmax(self, z):
        """
        Softmax activation function.

        Parameters:
        z (np.ndarray): Input array.

        Returns:
        np.ndarray: Output array after applying softmax.
        """
        exp_z = np.exp(z - np.max(z, axis=1, keepdims=True))
        return exp_z / np.sum(exp_z, axis=1, keepdims=True)

    def softmax_derivative(self, output, y):
        """
        Derivative of the softmax function for backpropagation.

        Parameters:
        output (np.ndarray): Output array from softmax function.
        y (np.ndarray): True labels.

        Returns:
        np.ndarray: Gradient of the loss with respect to softmax inputs.
        """
        return output - y

    def sigmoid(self, z):
        """
        Sigmoid activation function.

        Parameters:
        z (np.ndarray): Input array.

        Returns:
        np.ndarray: Output array after applying sigmoid.
        """
        return 1 / (1 + np.exp(-z))

    def sigmoid_derivative(self, z):
        """
        Derivative of the sigmoid function.

        Parameters:
        z (np.ndarray): Input array.

        Returns:
        np.ndarray: Derivative of sigmoid.
        """
        return z * (1 - z)

    def one_hot_encode(self, y, num_classes):
        """
        One hot encode the target variable.

        Parameters:
        y (np.ndarray): Target values.
        num_classes (int): Number of classes.

        Returns:
        np.ndarray: One hot encoded target values.
        """
        return np.eye(num_classes)[y]

    def forward(self, X):
        """
        Perform a forward pass through the network.

        Parameters:
        X (np.ndarray): Input data.

        Returns:
        list: List of activations for each layer.
        """
        activations = [X]
        for i in range(len(self.weights) - 1):
            z = np.dot(activations[-1], self.weights[i]) + self.biases[i]
            a = self.sigmoid(z)
            activations.append(a)
        # Output layer with softmax activation
        z = np.dot(activations[-1], self.weights[-1]) + self.biases[-1]
        a = self.softmax(z)
        activations.append(a)
        return activations

    def backward(self, X, y, activations):
        """
        Perform a backward pass and update weights and biases.

        Parameters:
        X (np.ndarray): Input data.
        y (np.ndarray): Target values.
        activations (list): List of activations for each layer.
        """
        m = X.shape[0]
        dz = self.softmax_derivative(activations[-1], y)
        for i in reversed(range(len(self.weights))):
            dw = np.dot(activations[i].T, dz) / m
            db = np.sum(dz, axis=0, keepdims=True) / m
            dz = np.dot(dz, self.weights[i].T) * self.sigmoid_derivative(activations[i])
            self.weights[i] -= self.learning_rate * dw
            self.biases[i] -= self.learning_rate * db

    def fit(self, X, y):
        """
        Train the neural network.

        Parameters:
        X (np.ndarray): Input data.
        y (np.ndarray): Target values.
        """
        y_encoded = self.one_hot_encode(y, self.layers[-1])
        for epoch in range(self.epochs):
            activations = self.forward(X)
            self.backward(X, y_encoded, activations)

    def predict(self, X):
        """
        Predict the output for given input data.

        Parameters:
        X (np.ndarray): Input data.

        Returns:
        np.ndarray: Predicted output.
        """
        activations = self.forward(X)
        return np.argmax(activations[-1], axis=1)

    def predict_proba(self, X):
        """
        Predict the probabilities for given input data.

        Parameters:
        X (np.ndarray): Input data.

        Returns:
        np.ndarray: Predicted probabilities.
        """
        activations = self.forward(X)
        return activations[-1]




from .roc_curve import roc_curve
from .auc import auc

def roc_auc_score(y_true, y_score):
    """Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC)
       from prediction scores.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        True binary labels in binary label indicators.

    y_score : array-like of shape (n_samples,)
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions
        (as returned by "decision_function" on some classifiers).

    Returns
    -------
    auc : float
        The Area Under the ROC Curve (ROC AUC) from prediction scores.
    """
    fpr, tpr, _ = roc_curve(y_true, y_score)
    return auc(fpr, tpr)





from metrics.regression import r2_score

class BaseEstimator:
    """Base class for all estimators in mini_sklearn."""

    def __init__(self):
        self._is_fitted = False
            
    def fit(self, X, y=None):
        
        """Fit estimator to data."""
        raise NotImplementedError("fit method is not implemented.")

    def score(self, X, y):
        """Return the coefficient of determination R^2 of the prediction."""
        raise NotImplementedError("score method is not implemented.")
    
    def check_is_fitted(self) -> bool:
        if not self._is_fitted:
            raise Exception('This estimator is not fitted.')

class TransformerMixin:
    """Mixin class for all transformers in mini_sklearn."""
    
    def fit_transform(self, X, y=None):
        self.fit(X)
        return self.transform(X)

class ClassifierMixin:
    """Mixin class for all classifiers in mini_sklearn."""
    
    def predict_proba(self, X):
        """Predict class probabilities."""
        raise NotImplementedError("predict_proba method is not implemented.")

    def predict_log_proba(self, X):
        """Predict class log-probabilities."""
        raise NotImplementedError("predict_log_proba method is not implemented.")

class RegressorMixin:
    """Mixin class for all regressors in mini_sklearn."""
    
    def score(self, X, y, sample_weight=None):
        """Return the coefficient of determination R^2 of the prediction."""
        y_pred = self.predict(X)

        return r2_score(y, y_pred)


from base.base import BaseEstimator, ClassifierMixin
import numpy as np
from scipy.special import logsumexp

class BaseNB(BaseEstimator, ClassifierMixin):
    def predict(self, X):
        self.check_is_fitted()

        joint_log_likelihood = self._joint_log_likelihood(X)
        return self.classes_[np.argmax(joint_log_likelihood, axis=1)]
    
    def predict_proba(self, X):
        self.check_is_fitted()

        joint_log_likelihood = self._joint_log_likelihood(X)
        log_prob = joint_log_likelihood - logsumexp(joint_log_likelihood, axis=1)[:, np.newaxis]
        return np.exp(log_prob)

    def _encode(self, y):
        classes = np.unique(y)
        y_train = np.zeros((y.shape[0], len(classes)))
        for i, c in enumerate(classes):
            y_train[y == c, i] = 1
        return classes, y_train

class GaussianNB(BaseNB):
    def fit(self, X, y):
        self.classes_ = np.unique(y)
        n_features = X.shape[1]
        n_classes = len(self.classes_)
        self.theta_ = np.zeros((n_classes, n_features))     # The means
        self.sigma_ = np.zeros((n_classes, n_features))     # The standard deviations
        self.class_count_ = np.zeros(n_classes)
        for i, c in enumerate(self.classes_):
            X_c = X[y == c]
            self.theta_[i] = np.mean(X_c, axis=0)
            self.sigma_[i] = np.var(X_c, axis=0)
            self.class_count_[i] = X_c.shape[0]
        self.class_prior_ = self.class_count_ / np.sum(self.class_count_)

        self._is_fitted = True
        return self

    def _joint_log_likelihood(self, X):
        self.check_is_fitted()

        joint_log_likelihood = np.zeros((X.shape[0], len(self.classes_)))
        for i in range(len(self.classes_)):
            p1 = np.log(self.class_prior_[i])
            p2 = -0.5 * np.log(2 * np.pi * self.sigma_[i]) - 0.5 * (X - self.theta_[i]) ** 2 / self.sigma_[i]
            joint_log_likelihood[:, i] = p1 + np.sum(p2, axis=1)
        return joint_log_likelihood


class BernoulliNB(BaseNB):
    def __init__(self, alpha=1.0):
        super().__init__()
        self.alpha = alpha

    def fit(self, X, y):        
        self.classes_, y_train = self._encode(y)
        self.feature_count_ = np.dot(y_train.T, X)
        self.class_count_ = y_train.sum(axis=0)
        smoothed_fc = self.feature_count_ + self.alpha
        smoothed_cc = self.class_count_ + 2 * self.alpha
        self.feature_log_prob_ = (np.log(smoothed_fc) -
                                  np.log(smoothed_cc.reshape(-1, 1)))
        self.class_log_prior_ = np.log(self.class_count_) - np.log(self.class_count_.sum())

        self._is_fitted = True
        return self
    
    def _joint_log_likelihood(self, X):
        self.check_is_fitted()

        return (np.dot(X, self.feature_log_prob_.T) +
                np.dot(1 - X, np.log(1 - np.exp(self.feature_log_prob_)).T) +
                self.class_log_prior_)
    

class MultinomialNB(BaseNB):
    def __init__(self, alpha=1.0):
        super().__init__()
        self.alpha = alpha

    def fit(self, X, y):
        self.classes_, y_train = self._encode(y)
        self.feature_count_ = np.dot(y_train.T, X)
        self.class_count_ = y_train.sum(axis=0)
        smoothed_fc = self.feature_count_ + self.alpha
        smoothed_cc = smoothed_fc.sum(axis=1)
        self.feature_log_prob_ = (np.log(smoothed_fc) -
                                  np.log(smoothed_cc.reshape(-1, 1)))
        self.class_log_prior_ = np.log(self.class_count_) - np.log(self.class_count_.sum())

        self._is_fitted = True
        return self

    def _joint_log_likelihood(self, X):
        self.check_is_fitted()

        return np.dot(X, self.feature_log_prob_.T) + self.class_log_prior_



from base.base import TransformerMixin, BaseEstimator
import numpy as np

class Normalizer(TransformerMixin, BaseEstimator):
    """
    Normalize samples individually to unit norm.

    This transformer normalizes each sample to have unit norm.
    The 'norm' parameter specifies the type of normalization to apply:
    - 'l1': L1 normalization (sum of absolute values)
    - 'l2': L2 normalization (Euclidean norm)
    - 'max': Max normalization (maximum absolute value)

    Parameters:
    -----------
    norm : str, default='l2'
        The normalization method to use. Options are 'l1', 'l2', or 'max'.

    Attributes:
    -----------
    norm : str
        The normalization method used.
    """

    def __init__(self, norm='l2'):
        if norm not in ['l1', 'l2', 'max']:
            raise ValueError('Invalid norm: %s' % norm)
        
        self.norm = norm

    def fit(self, X, y=None):
        """
        Fit the Normalizer to the data.

        Parameters:
        -----------
        X : array-like of shape (n_samples, n_features)
            The input data.

        Returns:
        --------
        self : object
            Returns self.
        """
        return self

    def transform(self, X):
        """
        Transform the input data by normalizing each sample.

        Parameters:
        -----------
        X : array-like of shape (n_samples, n_features)
            The input data.

        Returns:
        --------
        X_normalized : array-like of shape (n_samples, n_features)
            The normalized input data.
        """
        if self.norm == 'l1':
            norms = np.sum(np.abs(X), axis=1)
        elif self.norm == 'l2':
            norms = np.sqrt(np.sum(np.square(X), axis=1))
        elif self.norm == 'max':
            norms = np.max(np.abs(X), axis=1)
        return X / norms[:, np.newaxis]



import numpy as np
def check_X_y(X, y):
    """Validate and return X and y arrays."""
    X = np.asarray(X)
    y = np.asarray(y)
    if X.shape[0] != y.shape[0]:
        raise ValueError("Number of samples in X and y does not match.")
    return X, y


import numpy as np


def cosine_similarity(X, Y=None):
    # Normalize input data
    X_normalized = _normalize(X, copy=True)
    if Y is None:
        Y_normalized = X_normalized
    else:
        Y_normalized = _normalize(Y, copy=True)
    # Compute dot product
    dot_product = np.dot(X_normalized, Y_normalized.T)
    return dot_product


def _normalize(X, copy=False):
    X = np.asarray(X)
    if np.issubdtype(X.dtype, np.floating):
        norm = np.sqrt(np.sum(np.square(X), axis=1, keepdims=True))
        X_normalized = X / norm
        if copy:
            return X_normalized.copy()
        else:
            return X_normalized
    else:
        raise ValueError("Unsupported dtype for input array. Only floating-point arrays are supported.")



# mini_sklearn/utils/validation.py

import numpy as np
import warnings
import numbers


def _check_sample_weight(sample_weight, X):
    """Validate sample weights.

    Parameters
    ----------
    sample_weight : array-like of shape (n_samples,)
        Sample weights.

    X : array-like of shape (n_samples, n_features)
        Training data.

    Returns
    -------
    sample_weight : ndarray of shape (n_samples,)
        Validated sample weights.
    """
    if sample_weight is None:
        return np.ones(X.shape[0], dtype=np.float64)
    else:
        sample_weight = np.asarray(sample_weight)
        if np.any(sample_weight < 0):
            raise ValueError("Sample weights must be non-negative.")
        if len(sample_weight) != X.shape[0]:
            raise ValueError("Sample weights length must be equal to the number of samples.")
        return sample_weight



def _check_partial_fit_first_call(estimator, classes):
    """Check if partial_fit is called for the first time.

    Parameters
    ----------
    estimator : object
        Estimator instance.

    classes : array-like of shape (n_classes,)
        Unique class labels.

    Returns
    -------
    bool
        True if partial_fit is called for the first time, False otherwise.
    """
    if getattr(estimator, "classes_", None) is None and classes is None:
        raise ValueError(
            "classes must be passed on the first call to partial_fit."
        )
    return getattr(estimator, "classes_", None) is None


def column_or_1d(y, *, dtype=None, warn=False):
    """Convert input data to a 1d numpy array.

    Parameters
    ----------
    y : array-like
        Input data.

    dtype : data-type, default=None
        Data type for `y`.

    warn : bool, default=False
        Whether to display warnings.

    Returns
    -------
    y : ndarray
        Output data.

    Raises
    ------
    ValueError
        If `y` is not a 1D array or a 2D array with a single row or column.

    Examples
    --------
    >>> column_or_1d([1, 1])
    array([1, 1])
    """
    y = np.asarray(y, dtype=dtype)

    if y.ndim == 1:
        return y.ravel()
    elif y.ndim == 2 and (y.shape[0] == 1 or y.shape[1] == 1):
        if warn:
            warnings.warn("A column-vector y was passed when a 1d array was expected. Please change the shape of y to (n_samples, ).", UserWarning, stacklevel=2)
        return y.ravel()

    raise ValueError(f"y should be a 1d array, got an array of shape {y.shape} instead.")



def _num_samples(x):
    """Return number of samples in array-like x."""
    if hasattr(x, "__len__"):
        return len(x)
    elif hasattr(x, "shape") and x.shape is not None:
        if len(x.shape) == 0:
            raise TypeError("Singleton array %r cannot be considered a valid collection." % x)
        elif isinstance(x.shape[0], numbers.Integral):
            return x.shape[0]
    raise TypeError("Expected sequence or array-like, got %s" % type(x))



def check_array(array, *, ensure_2d=True, ensure_min_samples=1, ensure_min_features=1):
    if ensure_2d and array.ndim < 2:
        raise ValueError("Expected 2D array, got array with ndim=%d" % array.ndim)

    if ensure_min_samples > 0 and len(array) < ensure_min_samples:
        raise ValueError(
            "Found array with %d sample(s) while a minimum of %d is required"
            % (len(array), ensure_min_samples)
        )

    if ensure_min_features > 0 and array.ndim == 2 and array.shape[1] < ensure_min_features:
        raise ValueError(
            "Found array with %d feature(s) while a minimum of %d is required"
            % (array.shape[1], ensure_min_features)
        )

    return array



import numpy as np
from scipy.linalg import lstsq
from base.base import RegressorMixin, BaseEstimator

class LinearRegression(RegressorMixin, BaseEstimator):
    """
    Linear Regression model that uses the normal equation to fit and predict.

    Attributes:
        coef_ (np.ndarray): Coefficients of the linear regression model.
        intercept_ (float): Intercept of the linear regression model.
    """

    def fit(self, X, y):
        """
        Fit the linear regression model using the training data.

        Parameters:
            X (np.ndarray): Training data of shape (n_samples, n_features).
            y (np.ndarray): Target values of shape (n_samples,).

        Returns:
            self (LinearRegression): Returns the instance itself.
        """
        # Add a column of ones to X to account for the intercept
        X_with_intercept = np.hstack((np.ones((X.shape[0], 1)), X))
        
        # Solve the normal equation
        coef, _, _, _ = lstsq(X_with_intercept, y)
        
        # Store the coefficients and intercept
        self.intercept_ = coef[0]
        self.coef_ = coef[1:]

        self._is_fitted = True
        return self

    def predict(self, X):
        """
        Predict target values using the linear regression model.

        Parameters:
            X (np.ndarray): Input data of shape (n_samples, n_features).

        Returns:
            y_pred (np.ndarray): Predicted values of shape (n_samples,).
        """
        self.check_is_fitted()

        y_pred = np.dot(X, self.coef_) + self.intercept_
        
        return y_pred






import numpy as np
from .cosine_similarity import cosine_similarity


def cosine_distances(X, Y=None):
    # Compute cosine similarity
    similarity = cosine_similarity(X, Y)
    # Convert similarity to distances
    distances = 1.0 - similarity
    return distances



import warnings
warnings.filterwarnings("ignore")

import numpy as np
from sklearn.datasets import load_breast_cancer, load_diabetes
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, mean_squared_error
from sklearn.linear_model import LinearRegression as skLinearRegression
from sklearn.linear_model import LogisticRegression as skLogisticRegression

from linear_model.regression import LinearRegression
from linear_model.logistic_regression import LogisticRegression

# --- Linear Regression Test ---

# Load the diabetes dataset for regression
X, y = load_diabetes(return_X_y=True)

# Splitting the data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=100)

# Training my linear regression model
my_linear_reg = LinearRegression().fit(X_train, y_train)

# Training sklearn's linear regression model
sk_linear_reg = skLinearRegression().fit(X_train, y_train)

# Calculating mean squared error
mse_my_model = mean_squared_error(y_test, my_linear_reg.predict(X_test))
mse_sk_model = mean_squared_error(y_test, sk_linear_reg.predict(X_test))

# Printing the mean squared error
print("Mean Squared Error of my linear regression model:", mse_my_model)
print("Mean Squared Error of sklearn's linear regression model:", mse_sk_model)

print("--------------------------------------------------------------------------------")

# --- Logistic Regression Test ---

# Load the breast cancer dataset for classification
X, y = load_breast_cancer(return_X_y=True)

# Splitting the data into training and testing sets
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=100)

# Training my logistic regression model
my_logistic_reg = LogisticRegression().fit(X_train, y_train)

# Training sklearn's logistic regression model
sk_logistic_reg = skLogisticRegression().fit(X_train, y_train)

# Calculating accuracy scores
accuracy_my_model = accuracy_score(y_test, my_logistic_reg.predict(X_test))
accuracy_sk_model = accuracy_score(y_test, sk_logistic_reg.predict(X_test))

# Printing the accuracy scores
print("Accuracy score of my logistic regression model:", accuracy_my_model)
print("Accuracy score of sklearn's logistic regression model:", accuracy_sk_model)


from setuptools import setup, find_packages

setup(
    name="um6p_CC_learn",
    version="0.2.2",
    author="Sami AGOURRAM , Ilyass Hakkou",
    author_email="Sami.AGOURRAM@um6p.ma",
    description="A home made sklearn",
    packages=find_packages(),
    install_requires=[
        "numpy>=1.21.0",
        "scipy>=1.7.0",
        "scikit-learn>=0.24.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import numpy as np

def auc(x, y):
    """Compute Area Under the Curve (AUC) using the trapezoidal rule.

    Parameters
    ----------
    x : array-like of shape (n,)
        X coordinates. These must be either monotonic increasing or monotonic
        decreasing.
    y : array-like of shape (n,)
        Y coordinates.

    Returns
    -------
    auc : float
        Area Under the Curve.

    Raises
    ------
    ValueError
        If x is neither increasing nor decreasing.

    Notes
    -----
    This function assumes that x is sorted in increasing order.
    """
    # Check for consistent length of x and y
    if len(x) != len(y):
        raise ValueError("x and y must have the same length")

    # Check if x is monotonic
    if not np.all(np.diff(x) >= 0) and not np.all(np.diff(x) <= 0):
        raise ValueError("x must be monotonic increasing or decreasing")

    # Calculate the area under the curve using the trapezoidal rule
    auc_value = np.trapz(y, x)

    return auc_value



import numpy as np
from scipy.stats import mode

from base.base import TransformerMixin, BaseEstimator

class SimpleImputer(TransformerMixin, BaseEstimator):
    def __init__(self, strategy='mean', fill_value=None):
        self.strategy = strategy
        self.fill_value = fill_value  # only used when strategy == 'constant'

    def fit(self, X):
        if self.strategy not in ["mean", "median", "most_frequent", "constant"]:
            raise ValueError("Invalid strategy: %s. Strategy must be one of 'mean', 'median', 'most_frequent', or 'constant'." % self.strategy)
        
        if self.strategy == "constant" and self.fill_value is None:
            raise ValueError("Strategy is 'constant' but no fill value provided.")
        
        if self.strategy == "mean":
            self.statistics_ = np.nanmean(X, axis=0)
        elif self.strategy == "median":
            self.statistics_ = np.nanmedian(X, axis=0)
        elif self.strategy == "most_frequent":
            self.statistics_ = mode(X, axis=0, nan_policy='omit')[0].squeeze()
        elif self.strategy == "constant":
            self.statistics_ = np.full(X.shape[1], self.fill_value)
        self._is_fitted = True
        
        return self

    def transform(self, X):
        Xt = X.copy()
        missing_mask = np.isnan(Xt)
        if self.strategy != "constant":
            for i in range(Xt.shape[1]):
                missing = missing_mask[:, i]
                if np.any(missing):
                    Xt[missing, i] = self.statistics_[i]
        else:
            Xt[missing_mask] = self.fill_value
        
        return Xt



import numpy as np
from base.base import TransformerMixin, BaseEstimator

class StandardScaler(TransformerMixin, BaseEstimator):
    """Standardize features by removing the mean and scaling to unit variance.

    Standardization of a dataset is a common requirement for many machine learning estimators. 
    StandardScaler implements the Transformer API to compute the mean and standard deviation 
    on a training set so as to be able to later reapply the same transformation on the testing set.

    Parameters:
    None

    Attributes:
    mean_ : array of shape (n_features,)
        The mean value for each feature in the training set.
    scale_ : array of shape (n_features,)
        The standard deviation for each feature in the training set.

    Examples:
    >>> scaler = StandardScaler()
    >>> data = np.array([[0, 0], [0, 0], [1, 1], [1, 1]])
    >>> scaler.fit(data)
    StandardScaler()
    >>> scaler.mean_
    array([0.5, 0.5])
    >>> scaler.scale_
    array([0.5, 0.5])
    >>> scaler.transform(data)
    array([[-1., -1.],
           [-1., -1.],
           [ 1.,  1.],
           [ 1.,  1.]])
    """

    def fit(self, X, y=None):
        """Compute the mean and standard deviation for each feature in X.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Training data.

        Returns:
        self : StandardScaler
            The fitted StandardScaler instance.
        """
        self.mean_ = np.mean(X, axis=0)
        self.scale_ = np.std(X, axis=0)
        return self

    def transform(self, X):
        """Standardize features by removing the mean and scaling to unit variance.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Data to standardize.

        Returns:
        X_scaled : array-like of shape (n_samples, n_features)
            Transformed data.
        """
        return (X - self.mean_) / self.scale_
    


class MinMaxScaler(TransformerMixin, BaseEstimator):
    """Transforms features by scaling each feature to a given range.

    This scaler scales each feature to a given range. This range is determined
    by the `feature_range` parameter, which defaults to (0, 1).

    Parameters:
    feature_range : tuple, default=(0, 1)
        Desired range of transformed data.

    Attributes:
    data_min_ : array of shape (n_features,)
        The minimum value for each feature in the training set.
    data_max_ : array of shape (n_features,)
        The maximum value for each feature in the training set.
    data_range_ : array of shape (n_features,)
        The range of values (max - min) for each feature in the training set.
    scale_ : array of shape (n_features,)
        The scale factor applied to each feature.

    Examples:
    >>> scaler = MinMaxScaler()
    >>> data = np.array([[1, 2], [2, 3], [3, 4]])
    >>> scaler.fit(data)
    MinMaxScaler()
    >>> scaler.data_min_
    array([1, 2])
    >>> scaler.data_max_
    array([3, 4])
    >>> scaler.data_range_
    array([2, 2])
    >>> scaler.scale_
    array([0.5, 0.5])
    >>> scaler.transform(data)
    array([[0. , 0. ],
           [0.5, 0.5],
           [1. , 1. ]])
    """

    def __init__(self, feature_range=(0, 1)):
        self.feature_range = feature_range

    def fit(self, X, y=None):
        """Compute the minimum and maximum values for each feature in X.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Training data.

        Returns:
        self : MinMaxScaler
            The fitted MinMaxScaler instance.
        """
        self.data_min_ = X.min(axis=0)
        self.data_max_ = X.max(axis=0)
        self.data_range_ = self.data_max_ - self.data_min_
        self.scale_ = (self.feature_range[1] - self.feature_range[0]) / self.data_range_
        return self

    def transform(self, X):
        """Transform features by scaling each feature to the given range.

        Parameters:
        X : array-like of shape (n_samples, n_features)
            Data to scale.

        Returns:
        X_scaled : array-like of shape (n_samples, n_features)
            Transformed data.
        """
        return self.feature_range[0] + (X - self.data_min_) * self.scale_


import numpy as np

def euclidean_distances(X, Y=None, *, squared=False):
    """Compute Euclidean distances between samples in X and Y.

    Read more in the :ref:`User Guide <euclidean_distances>`.

    Parameters
    ----------
    X : {array-like, sparse matrix} of shape (n_samples_X, n_features)
        First set of samples.

    Y : {array-like, sparse matrix} of shape (n_samples_Y, n_features), \
            default=None
        Second set of samples. If `None`, the distances are computed between
        all pairs of samples in `X`.

    squared : bool, default=False
        Return squared Euclidean distances.

    Returns
    -------
    distances : ndarray of shape (n_samples_X, n_samples_Y) or \
                (n_samples_X, n_samples_X)
        Array containing the pairwise Euclidean distances. If `squared=True`,
        the array contains the squared distances.
    """
    # Check if Y is provided
    if Y is None:
        Y = X

    # Compute the pairwise Euclidean distances
    distances = -2 * np.dot(X, Y.T) + np.sum(X ** 2, axis=1)[:, np.newaxis] + np.sum(Y ** 2, axis=1)

    if not squared:
        # Clip distances to ensure they are non-negative
        distances = np.clip(distances, 0, None)
        distances = np.sqrt(distances)
    
    return distances




from utils._encode import _encode, _unique
from utils.__validation import column_or_1d, _num_samples

from base.base import BaseEstimator, TransformerMixin

import numpy as np



class LabelEncoder(TransformerMixin, BaseEstimator):
    """Encode target labels with value between 0 and n_classes-1.

    This transformer should be used to encode target values, *i.e.* `y`, and
    not the input `X`.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        Holds the label for each class.
    """

    def fit(self, y):
        """Fit label encoder.

        Parameters
        ----------
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        self : returns an instance of self.
            Fitted label encoder.
        """
        y = column_or_1d(y, warn=True)
        self.classes_ = _unique(y)
        self._is_fitted = True
        return self

    def fit_transform(self, y):
        """Fit label encoder and return encoded labels.

        Parameters
        ----------
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        y : array-like of shape (n_samples,)
            Encoded labels.
        """
        y = column_or_1d(y, warn=True)
        self.classes_, y = _unique(y, return_inverse=True)
        return y

    def transform(self, y):
        """Transform labels to normalized encoding.

        Parameters
        ----------
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        y : array-like of shape (n_samples,)
            Labels as normalized encodings.
        """
        self.check_is_fitted()
        y = column_or_1d(y, dtype=self.classes_.dtype, warn=True)
        # transform of empty array is empty array
        if _num_samples(y) == 0:
            return np.array([])

        return _encode(y, uniques=self.classes_)